#include "machine.h"
#include "cellspace.h"
#include <CL/cl_gl.h>
/**
Konstruktor ktory vytvory context s danymi zariadeniami
@param platform_index cislo platformy ktora sa ma pouzit
@param devices vektor zariadeni z ktorych bude vytvoreny kontext ak je prazdny pouziju sa vsetky zariadenia
*/
Machine::Machine(int platform_index, cl_device_v devices)
{
    cl_platform_v platform_ids = getPlatforms();

    cl_context_properties cl_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_ids[platform_index], 0 };
    init(cl_properties, devices);
    gl_interoperability = false;
}
/**
Konstruktor ktory vytvory context s danymi zariadeniami
@param gl_context OpenGL kontext s ktorym bude zviazany OpenCL kontext na sfunkcnenie OpenGL/OpenGL interoperability
@param display GLX display na ktorom je dany OpenGL zviazany
@param platform_index cislo platformy ktora sa ma pouzit
@param devices vektor zariadeni z ktorych bude vytvoreny kontext
*/
Machine::Machine(cl_context_properties gl_context, cl_context_properties display, int platform_index, cl_device_v devices)
{
    if(gl_context == 0 || display == 0)cerr << "Nie je inicializovane OpenGL" << endl;

    cl_platform_v platform_ids = getPlatforms();

    cl_context_properties cl_properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform_ids[platform_index],
        CL_GL_CONTEXT_KHR, gl_context, CL_GLX_DISPLAY_KHR, display, 0 };

    init(cl_properties, devices);
    gl_interoperability = true;
}
/**
pokracovanie konstruktora so spolocnym kodom.
@param cl_properties pole nastaveni privytvarani kontextu
@param devices vektor zariadeni ktore sa pouziju
*/
void Machine::init(cl_context_properties *cl_properties, cl_device_v &devices)
{
    cl_int err_code;
    if(devices.empty())context = clCreateContextFromType(cl_properties, CL_DEVICE_TYPE_ALL, NULL, NULL, &err_code);
    else context = clCreateContext(cl_properties, devices.size(), &devices[0], NULL, NULL, &err_code);

    if(err_code){ cerr << "Nepodarilo sa inicializovat OpenCL "; clGetError(err_code); return; }

    cl_char info[1000];
    cout << "Inicializujem OpenCL" << endl;
    clGetPlatformInfo((cl_platform_id)cl_properties[1], CL_PLATFORM_NAME, 1000, info, NULL); cout << "\tName:      \t" << info << endl;
    clGetPlatformInfo((cl_platform_id)cl_properties[1], CL_PLATFORM_VENDOR, 1000, info, NULL); cout << "\tVendor:  \t" << info << endl;
    clGetPlatformInfo((cl_platform_id)cl_properties[1], CL_PLATFORM_VERSION, 1000, info, NULL); cout << "\tVersion:\t" << info << endl;
    if(devices.empty())device = getDevices((cl_platform_id)cl_properties[1]);//ak je zoznam zariadeni prazny napln ho doplna
    else device = devices;
    for(cl_device_v::iterator i = device.begin(); i != device.end();i++)
    {
        queues.push_back(clCreateCommandQueue(context, *i, NULL, &err_code));
        if(err_code)
        {
            cerr << "Nepodarilo sa vytvorit OpenCL prikazove fronty "; clGetError(err_code);
            break;
        }
    }
    local_work_size = 64;
    deleting = false;
}
/**
destruktor uvolni vsetky prostriedky a taktiez aj svojich potomkov
*/
Machine::~Machine()
{
    for(vector<cl_command_queue>::iterator i = queues.begin(); i != queues.end(); i++)
        clReleaseCommandQueue(*i);
    deleting = true;//zablokujem odoberanie z potomkov
    for(vector<Model*>::iterator i =  models.begin(); i != models.end(); i++)
        delete *i;
    clReleaseContext(context);
}
/**
@return vektor obsahujuci vsetky platformy pristupne v systeme
*/
cl_platform_v Machine::getPlatforms()
{
    cl_int err_code;
    cl_platform_v platform_ids;
    cl_uint num_entries;
    err_code = clGetPlatformIDs(0, NULL, &num_entries);
    if(err_code)return platform_ids;
    platform_ids.resize(num_entries);
    err_code = clGetPlatformIDs(num_entries, &platform_ids[0], NULL);
    return platform_ids;
}
/**
@param platform_id ID platformy z ktorej sa maju vratit zariadenia
@param type typ zariadenia ktore ssa muju vratit CL_DEVICE_TYPE_GPU CL_DEVICE_TYPE_CPU
@return vektor zariadeni z danej platformy a typu
*/
cl_device_v Machine::getDevices(cl_platform_id platform_id, cl_device_type type)
{
    cl_device_v devices;
    cl_uint num_entries;
    cl_int err_code = clGetDeviceIDs(platform_id, type, 0, NULL, &num_entries);
    if(err_code)return devices;
    devices.resize(num_entries);
    clGetDeviceIDs(platform_id, type, num_entries, &devices[0], NULL);
    return devices;
}
/**
Registruje potomka.
@param child ukazovatel na potomka vola sa automaticky v konstruktore Model
*/
void Machine::addChild(Model *child)
{
    models.push_back(child);
}
/**
Odstrani potomka zo zoznau potomkov. Treba zmazat potom rucne. Volat z destruktora
@param child potomok ktory sa ma odstranit
*/
void Machine::removeChild(Model *child)
{
    if(!deleting)
        for(vector<Model*>::iterator i = models.begin(); i != models.end();i++)
            if(*i == child)
            {
                models.erase(i);
                break;
            }
}
/**
Prida do fronty na spustenie kernel aj s parametrami. Tie treba nastvit predtym ako sa zavola tato metoda.
Tato metoda nie je urcena na priame volanie z programu ale volanie z potomka.
@param kernel kernel ktory sa ma spustit
@param w sirka vypoctovej domeny
@param h vyska vypoctovej domeny
*/
void Machine::run(cl_kernel kernel, size_t w, size_t h, int device)
{
    if(device)device=0;
    else device=1;
    device=0;
    if(queues.size()==0)return;
    cl_int err_code;
    size_t global_work[] = { w, h };
    size_t local_work[] = { 16, 16 };
    err_code = clEnqueueNDRangeKernel(queues[device], kernel, 2, NULL, global_work, NULL, 0, NULL, NULL);
    if(err_code){ cerr << "Chyba pri spusteni kernela "; clGetError(err_code); }
}
/**
@return OpenCL kontext. Potrebne pre potomkov pri vytvarani programov a bufferov.
*/
cl_context Machine::getContext()
{
    return context;
}
/**
Zavola nad kazdou frontou clFinish co zabezpeci ze vsetky cakajuce udalosti sa vykonaju
*/
void Machine::finish()
{
    for(vector<cl_command_queue>::iterator i = queues.begin(); i != queues.end();i++)
        clFinish(*i);
}
/**
Skopiruje z CL buffera do pamete hosta
@param buf buffer z ktoreho sa bude kopirovat
@param data ukazovatel kam sa bude kopirovat
@param size velkost kopirovanych dat v bajtoch
@param block ak je true funkcia sa nevrati pokial nie su data skopirovane. ak je false tak sa kopirovanie vykona asynchronne
@return udalost ktora posluzi na synchronizaciu kopirovania. ak je blokove vrati NULL
*/
cl_event Machine::copy(cl_mem buf, void *data, size_t size, bool block)
{
    cl_int err_code;
    cl_event event = NULL;
    if(block)err_code = clEnqueueReadBuffer(queues[0], buf, CL_TRUE, 0, size, data, 0, NULL, NULL);
    else err_code = clEnqueueReadBuffer(queues[0], buf, CL_FALSE, 0, size, data, 0, NULL, &event);
    if(err_code){ cerr << "Chyba pri kopirovani z buffera"; clGetError(err_code); }
    return event;
}
/**
Skopiruje z CL buffera do pamete hosta
@param data ukazovatel odkial sa bude kopirovat
@param buf buffer do ktoreho sa bude kopirovat
@param size velkost kopirovanych dat v bajtoch
@param block ak je true funkcia sa nevrati pokial nie su data skopirovane. ak je false tak sa kopirovanie vykona asynchronne
@return udalost ktora posluzi na synchronizaciu kopirovania. ak je blokove vrati NULL
*/
cl_event Machine::copy(void *data, cl_mem buf, size_t size, bool block)
{
    cl_int err_code;
    cl_event event = NULL;
    if(block)err_code = clEnqueueWriteBuffer(queues[0], buf, CL_TRUE, 0, size, data, 0, NULL, NULL);
    else err_code = clEnqueueReadBuffer(queues[0], buf, CL_FALSE, 0, size, data, 0, NULL, &event);
    if(err_code){ cerr << "Chyba pri kopirovani do buffera"; clGetError(err_code); }
    return event;
}
/**
Skopiruje data medzi dvoma CL buffermi
@param buf1 buffer z ktoreho sa bude kopirovat
@param buf2 buffer do ktoreho sa bude kopirovat
@param size velkost kopirovanych dat v bajtoch
@param block ak je true funkcia vrati NULL. inak vracia event
@return udalost ktora posluzi na synchronizaciu kopirovania. ak je blokove vrati NULL
*/
cl_event Machine::copy(cl_mem buf1, cl_mem buf2, size_t size, bool block)
{
    cl_int err_code;
    cl_event event = NULL;
    if(block)err_code = clEnqueueCopyBuffer(queues[0], buf1, buf2, 0, 0, size, 0, NULL, NULL);
    else err_code = clEnqueueCopyBuffer(queues[0], buf1, buf2, 0, 0, size, 0, NULL, &event);
    if(err_code){ cerr << "Chyba pri kopirovani z buffera do buffera"; clGetError(err_code); }
    return event;
}
/**
Vytvory memory buffer na kontexte aktualnej Machine
@param flags standartne priznaky pre CLbuffer. CL_MEM_READ_ONLY,CL_MEM_WRITE_ONLY,CL_MEM_*_HOST_PTR
@param size velkost buffera v bajtoch
@param ptr ukazovatel na pouzivatelske data ak je pouzity CL_MEM_*_HOST_PTR priznak
*/
cl_mem Machine::createBuffer(cl_mem_flags flags, size_t size, void *ptr)
{
    cl_int err_code;
    cl_mem ret = clCreateBuffer(context, flags, size, ptr, &err_code);
    if(err_code){ cerr << "Nepodarilo sa vytvorit buffer"; clGetError(err_code); }
    return ret;
}
