#ifndef _MODEL_H_NOU_
#define _MODEL_H_NOU_

#include <string>
#include <vector>
#include <CL/cl.h>
#include "cellspace.h"

using namespace std;

class Machine;
/**
Abstraktna trieda poskytujuca jednotne rozhranie na pracu s roznymi modelmi.
Trieda ktora bude dedit musi implementovat tieto metody:
void run(CellSpace *pattern, int which, int iteration = 1);
int getCellSize();
void loadSource();
void renderToTexture(CellSpace *pattern, int which, cl_mem tex);
*/
class Model
{
    protected:
    cl_program program;
    vector<CellSpace*> patterns;
    public:
    Model(Machine *parent);
    virtual ~Model();
    Machine* getParent(){ return parent; }
    void addChild(CellSpace *child);
    void removeChild(CellSpace *child);
    /** Tato metoda by mala nastavit parametre pre kernel a následne zavolat parent->run()
    @param pattern ukazovatel na potomka obsahujuci stav modelu
    @param which vypocet bude prebiehat iterativne kedy sa zo stareho stavu vyratava novy. tento
        parameter urcuje ktory je ktory
    @param iteration kolko krat sa ma spustit kernel. pocas nich musi metoda preklapat which medzi 0 a 1
    */
    virtual void run(CellSpace *pattern, int which, int iteration = 1) = 0;
    /**
    @return pocet bajtov ktore zabera reprezentacia zdvojenie jednej bunky.
    */
    virtual int getCellSize() = 0;
    /**
    v tejto metode by sa mal vytvorit zdrojovy kod kernel a nasledne zavolat this->build() pre jeho
    skompletovanie
    */
    virtual void loadSource() = 0;
    /**
    Metoda by mala nastavit parametre kernel pre vykreslenie do buffera.
    @param pattern potom z ktoreho sa ma kreslit
    @param which ktory z dvojice sa ma vykreslit
    @param tex buffer do ktoreho sa bude kreslit
    */
    virtual void renderToTexture(CellSpace *pattern, int which, cl_mem tex) = 0;
    protected:
    cl_int build(string source, string options);
    cl_kernel kernel, render_kernel;
    Machine *parent;
};
/**
Trieda implemtujuca synchronne celularne automaty.
Su mozne dva druhy susedstva Neumanovo teda styria susedia hraniciaci s hranou alebo osem susedov
teda vsetky bunky ktore sa dotykaju strednej bunky.
Vstupnym formatom modelu je prechodovy strom vo forme matice s rozmermy pocet stavov krat pocet uzlov.
Kazdy riadok reprezentuje jeden uzol stromu a jednotlive stlpce nasledujuci uzol ktorym sa ma pokracovat.
*/
class SynchronousModel : public Model
{
    public:
    SynchronousModel(Machine *parent, string model_definition);
    int getCellSize(){ return sizeof(int)*2; }
    void loadSource();
    void run(CellSpace *pattern, int which, int iteration = 1);
    virtual void renderToTexture(CellSpace *pattern, int which, cl_mem tex);
    private:
    string source;
};

class AsynchronousModel : public Model
{
    private:
    int queue_len,neighbor;
    cl_kernel init_kernel,intermediate_kernel;
    public:
    AsynchronousModel(Machine* parent, string model_definition);
    ~AsynchronousModel();
    int getCellSize();
    void loadSource();
    void run(CellSpace *pattern, int which, int iteration = 1);
    void renderToTexture(CellSpace *pattern, int which, cl_mem tex);
    void initialize(CellSpace *pattern, float *data);
};

#endif // _MODEL_H_NOU_
